program lbm

  !--- modules --------------------------------------!
  use mod_precision
  use mod_sorties
  !--------------------------------------------------!

  implicit none

  !----signification des variables-------------------!
  !- reseau D2Q9
  ! qmax : nombre de vitesses
  ! c : vecteurs vitesse (adimensionnes) du reseau
  ! omega : poids
  ! vitesses "du son" : cd, cs
  ! cl : vecteur de meme taille que omega utilise pour les
  !      CLs avec l'obstacle

  !- maillage et pas de temps
  ! imax, jmax : nb de points dans les directions x et y
  ! x, y : abscisses et ordonnees des points du maillage
  ! dx : pas du maillage
  ! dt : pas de temps
  ! alpha : coefficient "cfl"
  ! obst : masque pour definir l'obstacle (pour le cas cylindre)

  !- dimensions du domaine
  ! L : longueur
  ! H : hauteur

  !- valeurs caracteristiques 
  ! xs : longueur
  ! us : vitesse
  ! rhos : densite

  !- parametres physiques
  ! mu : viscosite
  ! uw : vitesse de paroi (ou d'entree)
  ! Re : nb de Reynolds

  !- coefficient de relaxation
  ! tau : temps de relaxation 
  ! eta : coefficient du schema LBM

  !- quantites macroscopiques au temps tn et tn+1
  ! rho, rhonp1 : densite
  ! ux, uxnp1, uy, uynp1 : vitesse

  !- distributions
  ! gn : distribution au temps tn
  ! gnp1 : distribution au temps tn+1
  ! geq : distribution d'equilibre
  !--------------------------------------------------!
  

  !--- declarations ---------------------------------!
  integer, parameter :: qmax=8
  integer :: i,j,q,imax,jmax,iter
  integer, dimension(0:qmax,2) :: c
  integer, dimension(:,:), allocatable :: obst

  character(len=30) :: fich

  real(pr) :: L, H, rhos, mu, uw
  real(pr) :: dx, xs, us, Re
  real(pr) :: residu
  real(pr) :: cd, cs, tau, rhoZH
  real(pr) :: cfl, eta, nu, dt, alpha
  real(pr) :: temps
  real(pr) :: cx, cy, r

  real(pr), dimension(0:qmax) :: omega, cl
  real(pr), dimension(:), allocatable :: x, y
  real(pr), dimension(:,:), allocatable :: rho, ux, uy, rhonp1, uxnp1, uynp1
  real(pr), dimension(:,:,:), allocatable :: gn, gnp1, geq
  !--------------------------------------------------!


  !--- Debut du programme----------------------------!

  !--- lecture donnees 
  open(unit=12,file='data_couette.txt')
  read(12,*) L      ! longueur du domaine
  read(12,*) H      ! hauteur
  read(12,*) rhos   ! densite
  read(12,*) mu     ! viscosite
  read(12,*) uw    ! vitesse de le paroi
  read(12,*) alpha  ! coef cfl
  close(12)

  !--- lecture maillage
  open(unit=13,file='maillage.dat')
  read(13,*) jmax   ! nb mailles en y
  close(13)

  !--- calcul des parametres du schema
  !- pas du maillage et nb mailles en x
  dx = H/(jmax-1)
  imax =  nint((jmax-1)*L/H)+1

  !- pas de temps, vitesse du son, coef. relaxation
  nu = mu / rhos
  dt = alpha * dx**2 / (2*nu)
  cd = dx / dt
  cs = cd / sqrt(3._pr)
  tau = dt/2 + nu/cs**2
  eta = dt / tau

  !--- valeurs caracteristiques et Reynolds
  xs = H
  us = uw
  Re = rhos * us * xs / mu
  print*,'nombre de Reynolds base sur la largeur du canal : Re = ', Re

  !--- abscisses et ordonnees des noeuds du maillage 
  allocate(x(1:imax))
  do i=1,imax
     x(i) = (i-1)*dx
  end do

  allocate(y(1:jmax))
  do j=1,jmax
     y(j) = -H/2 + (j-1)*dx
  end do


  !--- definition de l'obstacle
  allocate(obst(1:imax,1:jmax))
  obst = 1  
!!$  cx = L/5._pr
!!$  cy = 0._pr
!!$  r = 0.2_pr*H/2._pr
!!$  obst = 1
!!$  do j=1,jmax
!!$     do i=1,imax
!!$
!!$        if ( (x(i)-cx)**2 + (y(j)-cy)**2 <= r**2 ) then
!!$           obst(i,j) = 0
!!$        end if
!!$
!!$     end do
!!$  end do
!!$  call sortie_obstacle(x,y,obst)


  !--- vitesses et poids
  omega(0) = 4._pr/9._pr
  omega(1:4) = 1._pr/9._pr
  omega(5:8) = 1._pr/36._pr

  !--- tableaux
  allocate(rho(1:imax,1:jmax),rhonp1(1:imax,1:jmax),ux(1:imax,1:jmax),uy(1:imax,1:jmax))
  allocate(geq(0:8,1:imax,1:jmax))
  allocate(uxnp1(1:imax,1:jmax),uynp1(1:imax,1:jmax))
  allocate(gn(0:8,0:imax+1,0:jmax+1))
  allocate(gnp1(0:8,0:imax+1,0:jmax+1))

  !--- vitesses discretes
  c(0,:) = (/0,0/)
  c(1,:) = (/1,0/)
  c(2,:) = (/0,1/)
  c(3,:) = (/-1,0/)
  c(4,:) = (/0,-1/)
  c(5,:) = (/1,1/)
  c(6,:) = (/-1,1/)
  c(7,:) = (/-1,-1/)
  c(8,:) = (/1,-1/)

  !--- initialisation
  !-- densite constante dans le domaine
  rho = rhos

  !-- vitesse adimensionnee (de la plaque pour Couette ou d'entree pour le cylindre)
  uw = uw / cd      
  print*,'vitesse uw/cd (doit etre <1 pour stab avec ZH)',uw
  if (uw>=1._pr) then
     stop
  end if

  !-- vitesse macro et distribution (changer la valeur du champ de vitesse pour le cylindre)
  do j=1,jmax
     do i=1,imax
        ux(i,j) = 0._pr
        uy(i,j) = 0._pr
        do q=0,8
           gn(q,i,j) = omega(q) *rho(i,j) * (                                             &
                &                        1._pr                                        &
                &                        + 3._pr * (ux(i,j)*c(q,1) + uy(i,j)*c(q,2))  &
                &                       )
        end do

     end do
  end do
  gnp1 = gn

  !--- boucle en temps
  iter = 0
  temps=0._pr
  residu = 1._pr
  boucle_temps:do while (residu>1.e-10_pr)

     iter = iter+1
     temps = temps + dt

     !--- collision (gn -> gn)
     !    (pour le cylindre : seulement  l'exterieur de l'obstacle)


     !--- transport (gn -> gnp1) 


     !--- conditions aux limites
     !- CL Ouest et Est : periodique

     !- CL Sud : paroi fixe (BB)

     !- CL Nord : paroi mobile (ZH)
     do i=1,imax
        j = jmax
        rhoZH = gnp1(0,i,j) + gnp1(1,i,j) + gnp1(3,i,j)           &
             &  + 2 * (gnp1(2,i,j) + gnp1(5,i,j) + gnp1(6,i,j))
        gnp1(4,i,j) = gnp1(2,i,j)
        gnp1(7,i,j) = -0.5_pr * (rhoZH*uw - gnp1(1,i,j) +gnp1(3,i,j) - 2._pr*gnp1(5,i,j))
        gnp1(8,i,j) =  0.5_pr * (rhoZH*uw - gnp1(1,i,j) + gnp1(3,i,j) + 2._pr*gnp1(6,i,j))
     end do

     !- CL Obstacle : uniquement pour le cas test du cylindre


     !--- calcul des quantites macroscopiques rhonp1, uxnp1, uynp1


     !--- calcul residu (uniquement pour Couette)
     residu = sum(abs(uxnp1-ux))/(sum(abs(uxnp1)))


     !--- mise a jour de gn, rho, ux, uy


     !--- sorties fichier des champs dimensionnels
     if (mod(iter,500)==0) then

        print*,'iter, temps=',iter,temps ,maxval(abs(rho))
        write(fich,*) iter
        fich='u'//trim(adjustl(fich))//'.vtk'
        call sorties_vtk(x,y,rho,ux*cd,uy*cd,obst,fich)

     end if

  end do boucle_temps

  !--- pour Couette : sortie du profil de vitesse (dimensionnee) au milieu du domaine
  open(unit=15,file='sol_lbm.dat')
  do j=1,jmax
     write(15,*) y(j), ux(imax/2,j) * cd
  end do
  close(15)

  !--- pour Couette : sortie du profil de vitesse exact (au milieu du domaine)
  open(unit=13,file='sol_ns.dat')
  dx = H/999
  do j=1,1000
     write(13,*) -H/2 + (j-1)*dx , uw*cd/H * (j-1)*dx
  end do
  close(13)

end program lbm
